#include "WiFi.h"

//////////////////////////////////////////////////////////////////////////
CWifi::CWifi()
{
	m_hNDUIO=NULL;
	InitializeCriticalSection(&m_Lock);
	OpenDriver();
}

//////////////////////////////////////////////////////////////////////////
CWifi::~CWifi()
{
	if(m_hNDUIO != NULL)
	{
		CloseDriver();
	}
	DeleteCriticalSection(&m_Lock);
}

//////////////////////////////////////////////////////////////////////////
//get list of usable adapters
bool CWifi::GetAdapters(LPWSTR pDest, DWORD &dwBufSizeBytes)
{
HANDLE hFile;
BYTE Buffer[2048];
void *pvBuf;
WCHAR TmpBuf[1024];
WCHAR *pszOut;
DWORD dwRet;
bool retval;

	if(pDest == NULL || dwBufSizeBytes < sizeof(WCHAR))
	{
		dwBufSizeBytes=0;
		return false;
	}

	EnterCriticalSection(&m_Lock);
	retval=false;
	//open NDIS driver
	hFile=CreateFile(L"NDS0:", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, INVALID_HANDLE_VALUE);
	if(hFile != INVALID_HANDLE_VALUE)
	{
		pvBuf=(void *)(&Buffer[0]);
		dwRet=sizeof(Buffer);
		if(DeviceIoControl(hFile, IOCTL_NDIS_GET_ADAPTER_NAMES, NULL, 0, pvBuf, sizeof(Buffer), &dwRet, NULL))
		{
			//adapter list ok.
			LPWSTR pszStr;
			dwRet=0;
			pszOut=TmpBuf;

			//no string classes used, so no MFC or ATL dependency.
			for(pszStr=(LPWSTR)pvBuf; *pszStr; pszStr+=wcslen(pszStr)+1) {
				//check if adapter name is ok, skip infrared, gprs, ActiveSync etc.
				if(wcsicmp(pszStr, L"ASYNCMAC1") && wcsicmp(pszStr, L"IRSIR1") &&
					wcsicmp(pszStr, L"L2TP1") && wcsicmp(pszStr, L"PPTP1") &&
					wcsicmp(pszStr, L"RNDISFN1") && wcsicmp(pszStr, L"WWAN1") &&
					wcsicmp(pszStr, L"XSC1_IRDA1")) {						
					//not the first adapter?
					if(pszOut != TmpBuf)
					{
						//append separator
						wcscat(pszOut, L",");
						pszOut++;
						dwRet+=sizeof(WCHAR);
					}
					wcscpy(pszOut, pszStr);
					pszOut+=wcslen(pszStr);
					dwRet+=sizeof(WCHAR) * wcslen(pszStr);
				}
			}
			//return required buffer size anyway
			dwBufSizeBytes=dwRet;

			//if buffer large enough, copy adapter names
			if(dwBufSizeBytes >= dwRet)
			{
				wcsncpy(pDest, TmpBuf, dwRet);
				retval=true;
			}
			//otherwise will return false
			//..
		}
		CloseHandle(hFile);
	}
	LeaveCriticalSection(&m_Lock);
	return retval;
}

//////////////////////////////////////////////////////////////////////////
//open the NDISUIO driver
bool CWifi::OpenDriver()
{
HANDLE hDev;

	//already opened?
	if(m_hNDUIO != NULL)
	{
		return true;
	}

	hDev=CreateFile(NDISUIO_DEVICE_NAME, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, INVALID_HANDLE_VALUE);
	if(hDev == INVALID_HANDLE_VALUE)
	{
		return false;
	}
	else
	{
		m_hNDUIO=hDev;
		return true;
	}
}

//////////////////////////////////////////////////////////////////////////
//at the end, close NDISUIO handle
void CWifi::CloseDriver()
{
	CloseHandle(m_hNDUIO);
	m_hNDUIO=NULL;
}

//////////////////////////////////////////////////////////////////////////
//initiate station scanning
bool CWifi::RefreshBSSIDs(LPWSTR pAdapter)
{
NDISUIO_SET_OID nso;
DWORD dwBytesRet;
bool retval;

	EnterCriticalSection(&m_Lock);
	nso.ptcDeviceName = pAdapter;
	nso.Oid = OID_802_11_BSSID_LIST_SCAN;

	dwBytesRet=0;
	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_SET_OID_VALUE, (void *)&nso, sizeof(NDISUIO_SET_OID), NULL, 0, &dwBytesRet, NULL))
	{
		retval=false;
	}
	else
	{
		retval=true;
	}
	LeaveCriticalSection(&m_Lock);
	return retval;
}

//////////////////////////////////////////////////////////////////////////
//get a list of currently visible stations
bool CWifi::GetBSSIDs(LPWSTR pAdapter, struct BSSIDInfo *pDest, DWORD &dwBufSizeBytes, DWORD &dwReturnedItems)
{
PNDISUIO_QUERY_OID pNQO;
DWORD dwBytesRet;
BYTE Buffer[8192], *pByte;
PNDIS_802_11_BSSID_LIST pList;
UINT i;

	if(pDest == NULL)
	{
		dwReturnedItems=0;
		return false;
	}

	EnterCriticalSection(&m_Lock);	
	pNQO=(PNDISUIO_QUERY_OID)Buffer;

	pNQO->ptcDeviceName = pAdapter;
	pNQO->Oid=OID_802_11_BSSID_LIST;

	// Run query
	dwBytesRet = 0;
	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_QUERY_OID_VALUE, (void *)pNQO, 8192, (void *)pNQO, 8192, &dwBytesRet, NULL))
	{
		LeaveCriticalSection(&m_Lock);
		return false;
	}

	pList=(PNDIS_802_11_BSSID_LIST)&pNQO->Data;
	dwReturnedItems=pList->NumberOfItems;

	//first item in array
	PNDIS_WLAN_BSSID pItem=pList->Bssid;

	for(i=0; i<pList->NumberOfItems; i++)
	{
		memcpy(pDest[i].BSSID, pItem->MacAddress, sizeof(pItem->MacAddress));
		mbstowcs(pDest[i].SSID, (char *)pItem->Ssid.Ssid, sizeof(pItem->Ssid.Ssid));
		pDest[i].RSSI=pItem->Rssi;
		if(pItem->Configuration.DSConfig > 14)
		{
			pDest[i].Channel=(pItem->Configuration.DSConfig - 2407000) / 5000;
		}
		else
		{
			pDest[i].Channel=pItem->Configuration.DSConfig;
		}
		pDest[i].Auth=pItem->Privacy;
		pDest[i].Infastructure=pItem->InfrastructureMode;

		//some pointer magic...actually pItem->Length was not sizeof(NDIS_WLAN_BSSID)
		//so use returned length
		pByte=(BYTE *)pItem;
		pByte+=pItem->Length;
		pItem=(PNDIS_WLAN_BSSID)pByte;
	}//for

	LeaveCriticalSection(&m_Lock);
	return true;
}

bool CWifi::SetAdHocConnection(LPWSTR pAdapter, LPSTR pSsid) {
	DWORD dwBytesRet;
	UCHAR SetBuffer[sizeof(NDISUIO_SET_OID) - sizeof(ULONG) + sizeof(NDIS_802_11_SSID)];
	PNDISUIO_SET_OID pSetOid;

	if(pAdapter == NULL || wcslen(pAdapter) == 0 || pSsid == NULL || strlen(pSsid) == 0) {
		return false;
	}

	EnterCriticalSection(&m_Lock);

	ZeroMemory(SetBuffer, sizeof(SetBuffer));
	pSetOid = (PNDISUIO_SET_OID)&SetBuffer[0];
	pSetOid->Oid = OID_802_11_SSID;

	PNDIS_802_11_SSID psSSID = (PNDIS_802_11_SSID)pSetOid->Data;
	psSSID->SsidLength = strlen((char *)pSsid);

	// L'SSID al massimo puo' essere di 31 byte
	if(psSSID->SsidLength >= 32)
		psSSID->SsidLength = 31;

	memcpy(psSSID->Ssid, pSsid, psSSID->SsidLength);	

	// Run query
	dwBytesRet = 0;
	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_SET_OID_VALUE, (void *)SetBuffer, sizeof(SetBuffer), 
						NULL, 0, &dwBytesRet, NULL)) {
		int err = GetLastError();
		LeaveCriticalSection(&m_Lock);
		return false;
	}
/*
	pList=(PNDIS_802_11_BSSID_LIST)&pNQO->Data;
	dwReturnedItems=pList->NumberOfItems;

	//first item in array
	PNDIS_WLAN_BSSID pItem=pList->Bssid;

	for(i=0; i<pList->NumberOfItems; i++)
	{
		memcpy(pDest[i].BSSID, pItem->MacAddress, sizeof(pItem->MacAddress));
		mbstowcs(pDest[i].SSID, (char *)pItem->Ssid.Ssid, sizeof(pItem->Ssid.Ssid));
		pDest[i].RSSI=pItem->Rssi;
		if(pItem->Configuration.DSConfig > 14)
		{
			pDest[i].Channel=(pItem->Configuration.DSConfig - 2407000) / 5000;
		}
		else
		{
			pDest[i].Channel=pItem->Configuration.DSConfig;
		}
		pDest[i].Auth=pItem->Privacy;
		pDest[i].Infastructure=pItem->InfrastructureMode;

		//some pointer magic...actually pItem->Length was not sizeof(NDIS_WLAN_BSSID)
		//so use returned length
		pByte=(BYTE *)pItem;
		pByte+=pItem->Length;
		pItem=(PNDIS_WLAN_BSSID)pByte;
	}//for
*/
	LeaveCriticalSection(&m_Lock);
	return true;
}

bool CWifi::SetConnectionMode(LPWSTR pAdapter, NDIS_802_11_NETWORK_INFRASTRUCTURE iMode, NDIS_802_11_AUTHENTICATION_MODE aMode) {
	NDISUIO_SET_OID sOid;
	DWORD dwBytesReturned;

	if(pAdapter == NULL || wcslen(pAdapter) == 0) {
		return false;
	}

	EnterCriticalSection(&m_Lock);

	sOid.ptcDeviceName = pAdapter;
	sOid.Oid = OID_802_11_AUTHENTICATION_MODE;
	memcpy(&sOid.Data, &aMode, sizeof(NDIS_802_11_AUTHENTICATION_MODE));

	// Settiamo l'authentication mode
	dwBytesReturned=0;
	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_SET_OID_VALUE, (void *)&sOid, sizeof(NDISUIO_SET_OID), NULL, 
						0, &dwBytesReturned, NULL)) {
		LeaveCriticalSection(&m_Lock);
		return false;
	}
			
	// E poi il tipo di infrastruttura
	sOid.Oid = OID_802_11_INFRASTRUCTURE_MODE;  
	memcpy(&sOid.Data, &iMode, sizeof(NDIS_802_11_NETWORK_INFRASTRUCTURE));  

	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_SET_OID_VALUE, (void *)&sOid, sizeof(NDISUIO_SET_OID), NULL, 
						0, &dwBytesReturned, NULL)) {
		LeaveCriticalSection(&m_Lock);
		return false;
	} 

	return true;
}

bool CWifi::SetSsid(LPWSTR pAdapter, LPCSTR pSsid) {
	BYTE SetBuffer[sizeof(NDISUIO_SET_OID) + sizeof(NDIS_802_11_SSID)];
	PNDISUIO_SET_OID psOid;
	NDIS_802_11_SSID ssid;
	DWORD dwBytesReturned, dwInFlags;

	if(pAdapter == NULL || wcslen(pAdapter) == 0 || pSsid == NULL || strlen(pSsid) == 0) {
		return false;
	}

	EnterCriticalSection(&m_Lock);

	ZeroMemory(&SetBuffer, sizeof(SetBuffer));

	psOid = (PNDISUIO_SET_OID)&SetBuffer;
	psOid->ptcDeviceName = pAdapter;
	psOid->Oid = OID_802_11_SSID;

	ssid.SsidLength = strlen(pSsid);

	if(ssid.SsidLength > 32)
		ssid.SsidLength = 32;

	memcpy(ssid.Ssid, pSsid, ssid.SsidLength);
	memcpy(psOid->Data, &ssid, sizeof(NDIS_802_11_SSID));

	// Impostiamo l'SSID
	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_SET_OID_VALUE, (void *)psOid, sizeof(SetBuffer), NULL, 
		0, &dwBytesReturned, NULL)) {
		LeaveCriticalSection(&m_Lock);
		return false;
	}

	INTF_ENTRY Intf;
	memset(&Intf, 0x00, sizeof(INTF_ENTRY));
	Intf.wszGuid = pAdapter;
	DWORD status = WZCSetInterface(NULL, INTF_PREFLIST, &Intf, &dwInFlags);

	LeaveCriticalSection(&m_Lock);
	return true;
}

bool CWifi::GetBSSID(LPWSTR pAdapter, LPWSTR pSsid, struct BSSIDInfo *pDest) {
	BSSIDInfo bi[20];
	DWORD len = sizeof(bi);
	DWORD retr, i;

	//
	INTF_ENTRY ientry;
	ZeroMemory(&ientry, sizeof(ientry));
	ientry.wszGuid = pAdapter;
	if(WZCSetInterface(NULL, INTFCTL_FALLBACK, &ientry, NULL) != ERROR_SUCCESS){
		return false;
	}

	BYTE SetBuffer[sizeof(NDISUIO_SET_OID) + sizeof(NDIS_802_11_SSID)];
	DWORD dwBytesReturned;
	PNDISUIO_SET_OID psOid;
	ZeroMemory(&SetBuffer, sizeof(SetBuffer));

	psOid = (PNDISUIO_SET_OID)&SetBuffer;
	psOid->ptcDeviceName = pAdapter;
	psOid->Oid = OID_802_11_DISASSOCIATE;

	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_SET_OID_VALUE, (void *)psOid, sizeof(SetBuffer), NULL, 
		0, &dwBytesReturned, NULL)) {
			LeaveCriticalSection(&m_Lock);
			return false;
	}
	//

	ZeroMemory(bi, sizeof(bi));

	if(pAdapter == NULL || wcslen(pAdapter) == 0 || pDest == NULL)
		return false;

	Sleep(2000); // Diamo alla scheda il tempo di refreshare le reti

	if(RefreshBSSIDs(pAdapter) == false)
		return false;

	if(GetBSSIDs(pAdapter, bi, len, retr) == false)
		return false;

	for(i = 0; i < retr; i++) {
		if(!memcmp(bi[i].SSID, pSsid, WideLen(pSsid))) {
			memcpy(pDest, &bi[i], sizeof(BSSIDInfo));
			return true;
		}
	}

	return false;
}

bool CWifi::SetWPAKey(LPWSTR pAdapter, struct BSSIDInfo *pBssid, BYTE* pKey, UINT uKeyLen) {
	PNDIS_802_11_WEP pWepKey = NULL;
	NDIS_802_11_ENCRYPTION_STATUS NdisEncryption;
	BYTE SetStatusBuffer[sizeof(NDISUIO_SET_OID) + sizeof(NDIS_802_11_ENCRYPTION_STATUS)];
	PNDISUIO_SET_OID psOid, psKeyOid = NULL;
	DWORD dwBytesReturned;

	if(pAdapter == NULL || wcslen(pAdapter) == 0 || pBssid == NULL || uKeyLen > 32)
		return false;

	EnterCriticalSection(&m_Lock);

	// 1. Impostiamo il tipo di cifratura
	ZeroMemory(&SetStatusBuffer, sizeof(SetStatusBuffer));

	psOid = (PNDISUIO_SET_OID)&SetStatusBuffer;
	psOid->ptcDeviceName = pAdapter;
	psOid->Oid = OID_802_11_ENCRYPTION_STATUS;

	NdisEncryption = Ndis802_11Encryption1Enabled;
	memcpy(psOid->Data, &NdisEncryption, sizeof(NDIS_802_11_ENCRYPTION_STATUS));

	// Settiamo l'encryption status
	dwBytesReturned=0;
	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_SET_OID_VALUE, (void *)psOid, sizeof(SetStatusBuffer), NULL, 
						0, &dwBytesReturned, NULL)) {

		LeaveCriticalSection(&m_Lock);
		return false;
	}

	// 2. Impostiamo la chiave
	psKeyOid = (PNDISUIO_SET_OID) new(std::nothrow) BYTE[sizeof(NDISUIO_SET_OID) + sizeof(NDIS_802_11_WEP) + uKeyLen];

	if(psKeyOid == NULL) {
		LeaveCriticalSection(&m_Lock);
		return false;
	}

	ZeroMemory(psKeyOid, sizeof(NDISUIO_SET_OID) + sizeof(NDIS_802_11_WEP) + uKeyLen);

	// Inizializza la struttura dell'OID
	psKeyOid->ptcDeviceName = pAdapter;
	psKeyOid->Oid = OID_802_11_ADD_WEP;

	// Puntiamo al membro Data[] della struttura
	pWepKey = (PNDIS_802_11_WEP)((BYTE *)psKeyOid + sizeof(NDISUIO_SET_OID) - sizeof(psKeyOid->Data));

	// Calcola la lunghezza della struttura
	pWepKey->Length = FIELD_OFFSET(NDIS_802_11_KEY, KeyMaterial) + uKeyLen;

	pWepKey->KeyIndex = 0x80000000;
	pWepKey->KeyLength = uKeyLen;

	// Copia la chiave nella struttura
	memcpy(pWepKey->KeyMaterial, pKey, pWepKey->KeyLength);

	if (!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_SET_OID_VALUE, (LPVOID)psKeyOid, sizeof(NDISUIO_SET_OID) + sizeof(NDIS_802_11_WEP) + uKeyLen,
						NULL, 0, &dwBytesReturned, NULL)) {

		delete[] psKeyOid;
		LeaveCriticalSection(&m_Lock);
		return false;
	}

	delete[] psKeyOid;

	// 3. Abilitiamo il WEP
	ZeroMemory(&SetStatusBuffer, sizeof(SetStatusBuffer));

	psOid = (PNDISUIO_SET_OID)&SetStatusBuffer;
	psOid->ptcDeviceName = pAdapter;
	psOid->Oid = OID_802_11_WEP_STATUS;

	NdisEncryption = Ndis802_11WEPEnabled;
	memcpy(psOid->Data, &NdisEncryption, sizeof(NDIS_802_11_ENCRYPTION_STATUS));

	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_SET_OID_VALUE, (void *)psOid, sizeof(SetStatusBuffer), NULL, 
						0, &dwBytesReturned, NULL)) {

		LeaveCriticalSection(&m_Lock);
		return false;
	}

	LeaveCriticalSection(&m_Lock);
	return true;
}

bool CWifi::CheckAssociation(LPWSTR pAdapter, LPCSTR pSsid) {
	PNDISUIO_QUERY_OID pNQO;
	PNDIS_802_11_ASSOCIATION_INFORMATION pInfo;
	BYTE SetBuffer[sizeof(NDISUIO_QUERY_OID) + 10*sizeof(NDIS_802_11_ASSOCIATION_INFORMATION)];
	DWORD dwBytesRet;

	if (pAdapter == NULL) {
		return false;
	}

	EnterCriticalSection(&m_Lock);	

	pNQO = (PNDISUIO_QUERY_OID)SetBuffer;
	pNQO->ptcDeviceName = pAdapter;
	pNQO->Oid = OID_802_11_ASSOCIATION_INFORMATION;

	// Run query
	dwBytesRet = 0;
	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDISUIO_QUERY_OID_VALUE, (void *)pNQO, sizeof(SetBuffer), (void *)pNQO, sizeof(SetBuffer), &dwBytesRet, NULL))
	{
		int err = GetLastError();
		LeaveCriticalSection(&m_Lock);
		return false;
	}

	pInfo = (PNDIS_802_11_ASSOCIATION_INFORMATION)(pNQO->Data);

	// richiedi i log
	dwBytesRet = 0;
	BYTE data[5000];
	WCHAR name[60];
	ZeroMemory(name, sizeof(name));
	wcscpy(name, L"TNETW12511");

	if(!DeviceIoControl(m_hNDUIO, IOCTL_NDIS_GET_LOG_DATA, (void *)name, wcslen(name)+1, data, sizeof(data), &dwBytesRet, NULL))
	{
		int err = GetLastError();
		LeaveCriticalSection(&m_Lock);
		return false;
	}

	return false;
}

bool CWifi::WZCConfigureNetwork(LPWSTR pAdapter, struct BSSIDInfo *pBssid, BYTE* pKey, UINT uKeyLen) {
	DWORD dwInFlags = 0;
	INTF_ENTRY Intf;
	memset(&Intf, 0x00, sizeof(INTF_ENTRY));
	Intf.wszGuid = pAdapter; // Your Adapter name

	WZC_WLAN_CONFIG wzcConfig1;
	memset(&wzcConfig1, 0, sizeof(wzcConfig1));

	wzcConfig1.Length = sizeof(wzcConfig1);

	// dwCtlFlags
	wzcConfig1.dwCtlFlags = 0;

	// Ssid
	WCHAR *szSsidToConnect = pBssid->SSID;

	wzcConfig1.Ssid.SsidLength = wcslen(szSsidToConnect);
	
	for(UINT i = 0; i<wzcConfig1.Ssid.SsidLength; i++)
		wzcConfig1.Ssid.Ssid[i] = (char)szSsidToConnect[i];

	// Privacy
	wzcConfig1.Privacy = Ndis802_11Encryption1Enabled;

	// adhoc? or infrastructure net?
	wzcConfig1.InfrastructureMode = Ndis802_11IBSS;

	// Authentication
	wzcConfig1.AuthenticationMode = Ndis802_11AuthModeShared;



	DWORD dwDataLen = sizeof(WZC_802_11_CONFIG_LIST);
	WZC_802_11_CONFIG_LIST *pNewConfigList = (WZC_802_11_CONFIG_LIST *)LocalAlloc(LPTR, dwDataLen);

	pNewConfigList->NumberOfItems = 1;
	pNewConfigList->Index = 0;

	memcpy(pNewConfigList->Config, &wzcConfig1, sizeof(wzcConfig1));
	Intf.rdStSSIDList.pData = (BYTE*)pNewConfigList;
	Intf.rdStSSIDList.dwDataLen = dwDataLen;

	DWORD dwStatus = WZCSetInterface(NULL, INTF_PREFLIST, &Intf, &dwInFlags);

	if(dwStatus)
		wprintf(L"WZCSetInterfaceEx() error dwStatus=0x%0X,	dwOutFlags=0x%0X", dwStatus, dwInFlags);
	else
		wprintf(L"WZC sets [Preferred Network]\n");

	// Free memory block that WZC allocated.
	// WZCQueryInterfaceEx() always should be followed by this	
	// Note that the wzcGuid is ** NOT ** freed by WZCDeleteIntfObj()
	WZCDeleteIntfObj(&Intf);

	LocalFree(pNewConfigList);

	return true;
}